/*
Copyright (c) 2013, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "Python.h"
#include "pyunconfig.h"

#include <grub/cpu/io.h>
#include <grub/datetime.h>
#include <grub/disk.h>
#include <grub/env.h>
#include <grub/partition.h>
#include <grub/term.h>
#include <grub/time.h>

#include "acpica.h"
#include "bitsmodule.h"
#include "smp.h"

struct dword_regs {
    U32 eax;
    U32 ebx;
    U32 ecx;
    U32 edx;
};

struct memop {
    U32 addr;
    U64 value;
};

static PyObject *bits_bclk(PyObject *self, PyObject *args)
{
    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    return Py_BuildValue("I", smp_read_bclk());
}

static U32 bsp_apicid(void) {
    const CPU_INFO *cpu;
    cpu = smp_read_cpu_list();
    return cpu[0].apicid;
}

static void blocking_sleep_callback(void *param)
{
    U32 *usec = param;
    smp_sleep(*usec);
}

static PyObject *bits_blocking_sleep(PyObject *self, PyObject *args)
{
    U32 usec;

    if (!PyArg_ParseTuple(args, "I", &usec))
        return NULL;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");

    smp_function(smp_read_cpu_list()[0].apicid, blocking_sleep_callback, &usec);
    return Py_BuildValue("");
}

static void cpuid_callback(void *param)
{
    struct dword_regs *r = param;
    cpuid32_indexed(r->eax, r->ecx, &r->eax, &r->ebx, &r->ecx, &r->edx);
}

/* Returns true for success, false for failure. */
static bool smp_cpuid(U32 apicid, struct dword_regs *regs)
{
    return !!smp_function(apicid, cpuid_callback, regs);
}

static PyObject *bits_cpuid(PyObject *self, PyObject *args)
{
    struct dword_regs regs = { .ecx=0 };
    unsigned apicid;
    int ncpus;

    if (!PyArg_ParseTuple(args, "II|I", &apicid, &regs.eax, &regs.ecx))
        return NULL;

    ncpus = smp_init();
    if (!ncpus)
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");

    if (!smp_cpuid(apicid, &regs))
        return PyErr_Format(PyExc_RuntimeError, "SMP function returned an error; does apicid 0x%x exist?", apicid);
    return Py_BuildValue("IIII", regs.eax, regs.ebx, regs.ecx, regs.edx);
}

static PyObject *bits_cpus(PyObject *self, PyObject *args)
{
    int ncpus;
    int ndx;
    PyObject *apicid_list;
    const CPU_INFO *cpu;

    ncpus = smp_init();
    if (!ncpus)
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");

    apicid_list = PyList_New(ncpus);
    cpu = smp_read_cpu_list();
    for (ndx = 0; ndx < ncpus; ndx++)
        PyList_SET_ITEM(apicid_list, ndx, Py_BuildValue("i", cpu[ndx].apicid));
    return Py_BuildValue("N", apicid_list);
}

struct msr {
    U32 num;
    U32 status;
    U64 value;
};

static void rdmsr_callback(void *param)
{
    struct msr *m = param;
    rdmsr64(m->num, &m->value, &m->status);
}

/* Returns true for success, false for failure. */
static bool smp_rdmsr(U32 apicid, struct msr *msr)
{
    msr->status = -1;
    smp_function(apicid, rdmsr_callback, msr);
    return !msr->status;
}

static PyObject *bits_rdmsr(PyObject *self, PyObject *args)
{
    struct msr msr;
    unsigned apicid;
    int ncpus;

    if (!PyArg_ParseTuple(args, "II", &apicid, &msr.num))
        return NULL;

    ncpus = smp_init();
    if (!ncpus)
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");

    if (!smp_rdmsr(apicid, &msr))
        return Py_BuildValue("");
    return Py_BuildValue("K", msr.value);
}

static void wrmsr_callback(void *param)
{
    struct msr *m = param;
    wrmsr64(m->num, m->value, &m->status);
}

/* Returns true for success, false for failure. */
static bool smp_wrmsr(U32 apicid, struct msr *msr)
{
    msr->status = -1;
    smp_function(apicid, wrmsr_callback, msr);
    return !msr->status;
}

static PyObject *bits_wrmsr(PyObject *self, PyObject *args)
{
    struct msr msr;
    unsigned apicid;
    int ncpus;

    if (!PyArg_ParseTuple(args, "IIK", &apicid, &msr.num, &msr.value))
        return NULL;

    ncpus = smp_init();
    if (!ncpus)
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");

    if (!smp_wrmsr(apicid, &msr))
        return Py_BuildValue("");
    return PyBool_FromLong(!msr.status);
}

static void inb_callback(void *param)
{
    struct memop *m = param;
    m->value = grub_inb(m->addr);
}

static char *in_keywords[] = {"port", "apicid", NULL};

static PyObject *bits_inb(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    U16 port;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "H|I", in_keywords, &port, &apicid))
        return NULL;
    m.addr = port;

    if (!smp_function(apicid, inb_callback, &m))
        return Py_BuildValue("");
    return Py_BuildValue("B", (U8)m.value);
}

static void inw_callback(void *param)
{
    struct memop *m = param;
    m->value = grub_inw(m->addr);
}

static PyObject *bits_inw(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    U16 port;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "H|I", in_keywords, &port, &apicid))
        return NULL;
    m.addr = port;

    if (!smp_function(apicid, inw_callback, &m))
        return Py_BuildValue("");
    return Py_BuildValue("H", (U16)m.value);
}

static void inl_callback(void *param)
{
    struct memop *m = param;
    m->value = grub_inl(m->addr);
}

static PyObject *bits_inl(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    U16 port;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "H|I", in_keywords, &port, &apicid))
        return NULL;
    m.addr = port;

    if (!smp_function(apicid, inl_callback, &m))
        return Py_BuildValue("");
    return Py_BuildValue("I", (U32)m.value);
}

static void outb_callback(void *param)
{
    struct memop *m = param;
    grub_outb(m->value, m->addr);
}

static char *out_keywords[] = {"port", "value", "apicid", NULL};

static PyObject *bits_outb(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    U16 port;
    U8 value;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "HB|I", out_keywords, &port, &value, &apicid))
        return NULL;
    m.addr = port;
    m.value = value;

    smp_function(apicid, outb_callback, &m);
    return Py_BuildValue("");
}

static void outw_callback(void *param)
{
    struct memop *m = param;
    grub_outw(m->value, m->addr);
}

static PyObject *bits_outw(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    U16 port;
    U16 value;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "HH|I", out_keywords, &port, &value, &apicid))
        return NULL;
    m.addr = port;
    m.value = value;

    smp_function(apicid, outw_callback, &m);
    return Py_BuildValue("");
}

static void outl_callback(void *param)
{
    struct memop *m = param;
    grub_outl(m->value, m->addr);
}

static PyObject *bits_outl(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    U16 port;
    U32 value;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "HI|I", out_keywords, &port, &value, &apicid))
        return NULL;
    m.addr = port;
    m.value = value;

    smp_function(apicid, outl_callback, &m);
    return Py_BuildValue("");
}

static void readb_callback(void *param)
{
    struct memop *m = param;
    m->value = *(U8 *)(m->addr);
}

static char *read_keywords[] = {"address", "apicid", NULL};

static PyObject *bits_readb(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "I|I", read_keywords, &m.addr, &apicid))
        return NULL;

    if (!smp_function(apicid, readb_callback, &m))
        return Py_BuildValue("");
    return Py_BuildValue("B", (U8)m.value);
}

static void readw_callback(void *param)
{
    struct memop *m = param;
    m->value = *(U16 *)(m->addr);
}

static PyObject *bits_readw(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "I|I", read_keywords, &m.addr, &apicid))
        return NULL;

    if (!smp_function(apicid, readw_callback, &m))
        return Py_BuildValue("");
    return Py_BuildValue("H", (U16)m.value);
}

static void readl_callback(void *param)
{
    struct memop *m = param;
    m->value = *(U32 *)(m->addr);
}

static PyObject *bits_readl(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "I|I", read_keywords, &m.addr, &apicid))
        return NULL;

    if (!smp_function(apicid, readl_callback, &m))
        return Py_BuildValue("");
    return Py_BuildValue("I", (U32)m.value);
}

static void readq_callback(void *param)
{
    struct memop *m = param;
    m->value = *(U64 *)m->addr;
}

static PyObject *bits_readq(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "I|I", read_keywords, &m.addr, &apicid))
        return NULL;

    if (!smp_function(apicid, readq_callback, &m))
        return Py_BuildValue("");
    return Py_BuildValue("K", m.value);
}

static void writeb_callback(void *param)
{
    struct memop *m = param;
    *(U8 *)(m->addr) = m->value;
}

static char *write_keywords[] = {"address", "value", "apicid", NULL};

static PyObject *bits_writeb(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    U8 value;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "IB|I", write_keywords, &m.addr, &value, &apicid))
        return NULL;
    m.value = value;

    smp_function(apicid, writeb_callback, &m);
    return Py_BuildValue("");
}

static void writew_callback(void *param)
{
    struct memop *m = param;
    *(U16 *)(m->addr) = m->value;
}

static PyObject *bits_writew(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    U16 value;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "IH|I", write_keywords, &m.addr, &value, &apicid))
        return NULL;
    m.value = value;

    smp_function(apicid, writew_callback, &m);
    return Py_BuildValue("");
}

static void writel_callback(void *param)
{
    struct memop *m = param;
    *(U32 *)(m->addr) = m->value;
}

static PyObject *bits_writel(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    U32 value;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "II|I", write_keywords, &m.addr, &value, &apicid))
        return NULL;
    m.value = value;

    smp_function(apicid, writel_callback, &m);
    return Py_BuildValue("");
}

static void writeq_callback(void *param)
{
    struct memop *m = param;
    *(U64 *)(m->addr) = m->value;
}

static PyObject *bits_writeq(PyObject *self, PyObject *args, PyObject *keywds)
{
    struct memop m;
    unsigned apicid;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    apicid = bsp_apicid();

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "IK|I", write_keywords, &m.addr, &m.value, &apicid))
        return NULL;

    smp_function(apicid, writeq_callback, &m);
    return Py_BuildValue("");
}

static PyObject *pyblocklist = NULL;
static grub_disk_addr_t partition_start_sector = 0;

static void NESTED_FUNC_ATTR disk_blocks_read_hook(grub_disk_addr_t sector, unsigned offset, unsigned length)
{
    PyObject *tuple;
    if (!pyblocklist)
        return;
    tuple = Py_BuildValue("(KII)", sector - partition_start_sector, offset, length);
    if (!tuple || PyList_Append(pyblocklist, tuple) == -1)
        Py_CLEAR(pyblocklist);
    Py_XDECREF(tuple);
}

static PyObject *bits_file_data_and_disk_blocks(PyObject *self, PyObject *args)
{
    PyObject *pyfile, *pystr;
    grub_file_t file;
    grub_ssize_t bytes_read;

    if (!PyArg_ParseTuple(args, "O!:file_data_and_disk_blocks", &PyFile_Type, &pyfile))
        return NULL;

    file = PyFile_AsFile(pyfile);
    if (!file->device->disk)
        return PyErr_Format(PyExc_RuntimeError, "Can't get disk blocks from non-disk-backed file");

    partition_start_sector = grub_partition_get_start(file->device->disk->partition);

    pyblocklist = PyList_New(0);
    if (!pyblocklist)
        return NULL;
    pystr = PyString_FromStringAndSize(NULL, grub_file_size(file));
    if (!pystr) {
        Py_CLEAR(pyblocklist);
        return NULL;
    }

    file->read_hook = disk_blocks_read_hook;
    bytes_read = grub_file_read(file, PyString_AsString(pystr), grub_file_size(file));
    file->read_hook = NULL;
    if ((grub_off_t)bytes_read != grub_file_size(file)) {
        Py_CLEAR(pyblocklist);
        Py_DECREF(pystr);
        return PyErr_Format(PyExc_RuntimeError, "Failed to read from file");
    }
    if (!pyblocklist) {
        Py_DECREF(pystr);
        return PyErr_NoMemory();
    }

    return Py_BuildValue("(NN)", pystr, pyblocklist);
}

static PyObject *bits_disk_read(PyObject *self, PyObject *args)
{
    PyObject *pyfile, *pystr;
    grub_file_t file;
    grub_disk_addr_t sector;
    unsigned offset, length;

    if (!PyArg_ParseTuple(args, "O!KII:disk_read", &PyFile_Type, &pyfile, &sector, &offset, &length))
        return NULL;

    file = PyFile_AsFile(pyfile);
    if (!file->device->disk)
        return PyErr_Format(PyExc_RuntimeError, "Can't get disk device from non-disk-backed file");

    pystr = PyString_FromStringAndSize(NULL, length);
    if (!pystr)
        return PyErr_NoMemory();

    if (grub_disk_read(file->device->disk, sector, offset, length, PyString_AsString(pystr)) != GRUB_ERR_NONE) {
        Py_DECREF(pystr);
        return PyErr_SetFromErrno(PyExc_IOError);
    }

    return pystr;
}

static PyObject *bits_disk_write(PyObject *self, PyObject *args)
{
    PyObject *pyfile;
    grub_file_t file;
    grub_disk_addr_t sector;
    const char *data;
    unsigned offset;
    int length;

    if (!PyArg_ParseTuple(args, "O!KIs#:disk_write", &PyFile_Type, &pyfile, &sector, &offset, &data, &length))
        return NULL;

    file = PyFile_AsFile(pyfile);
    if (!file->device->disk)
        return PyErr_Format(PyExc_RuntimeError, "Can't get disk device from non-disk-backed file");

    if (grub_disk_write(file->device->disk, sector, offset, length, data) != GRUB_ERR_NONE)
        return PyErr_SetFromErrno(PyExc_IOError);

    return Py_BuildValue("");
}

static PyObject *acpi_object_to_python(ACPI_OBJECT *obj)
{
    if (obj == NULL)
        return Py_BuildValue("");

    switch(obj->Type) {
    case ACPI_TYPE_ANY:
        return Py_BuildValue("");
    case ACPI_TYPE_INTEGER:
        return Py_BuildValue("IK", ACPI_TYPE_INTEGER, obj->Integer.Value);
    case ACPI_TYPE_STRING:
        return Py_BuildValue("Is#", ACPI_TYPE_STRING, obj->String.Pointer, obj->String.Length);
    case ACPI_TYPE_BUFFER:
        return Py_BuildValue("Is#", ACPI_TYPE_BUFFER, obj->Buffer.Pointer, obj->Buffer.Length);
    case ACPI_TYPE_PACKAGE:
        {
            U32 ndx;
            PyObject *pkg = PyTuple_New(obj->Package.Count);
            for (ndx = 0; ndx < obj->Package.Count; ndx++)
                PyTuple_SET_ITEM(pkg, ndx, acpi_object_to_python(&obj->Package.Elements[ndx]));
            return Py_BuildValue("IN", ACPI_TYPE_PACKAGE, pkg);
        }
    case ACPI_TYPE_POWER:
        return Py_BuildValue("I(II)", ACPI_TYPE_POWER, obj->PowerResource.SystemLevel, obj->PowerResource.ResourceOrder);
    case ACPI_TYPE_PROCESSOR:
        return Py_BuildValue("I(III)", ACPI_TYPE_PROCESSOR, obj->Processor.ProcId, obj->Processor.PblkAddress, obj->Processor.PblkLength);
    default:
        return PyErr_Format(PyExc_RuntimeError,
                            "Unable to parse the ACPI object returned from acpi_object_to_python on unhandled ACPI_OBJECT_TYPE %u.",
                            obj->Type);
    }
}

static void free_acpi_objects(ACPI_OBJECT *objs, U32 count)
{
    if (objs) {
        U32 i;
        for (i = 0; i < count; i++)
            if (objs[i].Type == ACPI_TYPE_PACKAGE)
                free_acpi_objects(objs[i].Package.Elements, objs[i].Package.Count);
        grub_free(objs);
    }
}

static bool acpi_objects_from_python(PyObject *pyobj, ACPI_OBJECT **objs, U32 *count);

static bool acpi_object_from_python(PyObject *pyobj, ACPI_OBJECT *obj)
{
    PyObject *value;

    if (pyobj == Py_None) {
        obj->Type = ACPI_TYPE_ANY;
        return true;
    }
    if (!PyArg_ParseTuple(pyobj, "IO:acpi_object_from_python", &obj->Type, &value))
        return false;
    switch (obj->Type) {
    case ACPI_TYPE_INTEGER:
        obj->Integer.Value = PyInt_AsUnsignedLongLongMask(value);
        return true;
    case ACPI_TYPE_STRING:
        {
            Py_ssize_t length;
            if (PyString_AsStringAndSize(value, &obj->String.Pointer, &length) < 0)
                return false;
            obj->String.Length = length;
            return true;
        }
    case ACPI_TYPE_BUFFER:
        {
            Py_ssize_t length;
            if (PyString_AsStringAndSize(value, (char **)&obj->Buffer.Pointer, &length) < 0)
                return false;
            obj->Buffer.Length = length;
            return true;
        }
    case ACPI_TYPE_PACKAGE:
        return acpi_objects_from_python(value, &obj->Package.Elements, &obj->Package.Count);
    case ACPI_TYPE_POWER:
        return PyArg_ParseTuple(value, "II", &obj->PowerResource.SystemLevel, &obj->PowerResource.ResourceOrder);
    case ACPI_TYPE_PROCESSOR:
        return PyArg_ParseTuple(value, "III", &obj->Processor.ProcId, &obj->Processor.PblkAddress, &obj->Processor.PblkLength);
    default:
        PyErr_Format(PyExc_RuntimeError,
                     "Python object provided as ACPI method parameter used unhandled ACPI_OBJECT_TYPE %u.",
                     obj->Type);
        return false;
    }
}

static bool acpi_objects_from_python(PyObject *pyobj, ACPI_OBJECT **objs, U32 *count)
{
    U32 i;

    if (!PyTuple_Check(pyobj))
        return false;

    *count = PyTuple_Size(pyobj);
    if (*count) {
        *objs = grub_zalloc(*count * sizeof(**objs));
        if (!*objs)
            return false;
    } else
        *objs = NULL;

    for (i = 0; i < *count; i++) {
        if (!acpi_object_from_python(PyTuple_GetItem(pyobj, i), &(*objs)[i])) {
            free_acpi_objects(*objs, *count);
            return false;
        }
    }

    return true;
}

static PyObject *bits_acpi_eval(PyObject *self, PyObject *args)
{
    char *pathname;
    PyObject *acpi_args_tuple;
    ACPI_OBJECT_LIST acpi_args;
    ACPI_BUFFER results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    PyObject *ret;

    if (!PyArg_ParseTuple(args, "sO", &pathname, &acpi_args_tuple))
        return NULL;

    if (!acpi_objects_from_python(acpi_args_tuple, &acpi_args.Pointer, &acpi_args.Count))
        return NULL;

    if (acpica_init() != GRUB_ERR_NONE) {
        free_acpi_objects(acpi_args.Pointer, acpi_args.Count);
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");
    }

    if (ACPI_FAILURE(AcpiEvaluateObject(NULL, pathname, &acpi_args, &results))) {
        free_acpi_objects(acpi_args.Pointer, acpi_args.Count);
        return Py_BuildValue("");
    }

    free_acpi_objects(acpi_args.Pointer, acpi_args.Count);
    ret = acpi_object_to_python(results.Pointer);

    ACPI_FREE(results.Pointer);
    return Py_BuildValue("N", ret);
}

static PyObject *bits_acpi_get_object_info(PyObject *self, PyObject *args)
{
    char *pathname;
    ACPI_HANDLE handle = 0;
    ACPI_DEVICE_INFO *info = NULL;
    PyObject *ret;
    ACPI_DEVICE_INFO *addr;

    if (!PyArg_ParseTuple(args, "s", &pathname))
        return NULL;

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    if (ACPI_FAILURE(AcpiGetHandle(NULL, pathname, &handle)) || (handle == 0))
        return PyErr_Format(PyExc_RuntimeError, "Couldn't get object handle for \"%s\"", pathname);

    if (ACPI_FAILURE(AcpiGetObjectInfo(handle, &info)))
        return PyErr_Format(PyExc_RuntimeError, "Couldn't get object info for \"%s\"", pathname);

    ret = Py_BuildValue("(s#I)", info, info->InfoSize, info);
    ACPI_FREE(info);
    return ret;
}

static PyObject *bits_acpi_get_table(PyObject *self, PyObject *args)
{
    char *signature;
    U32 instance = 1;
    ACPI_TABLE_HEADER *table_header;

    if (!PyArg_ParseTuple(args, "s|I", &signature, &instance))
        return NULL;

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    if (ACPI_FAILURE(AcpiGetTable(signature, instance, &table_header)))
        return Py_BuildValue("");

    return Py_BuildValue("s#", table_header, table_header->Length);
}

static PyObject *bits_acpi_get_rsdp(PyObject *self, PyObject *args)
{
    ACPI_TABLE_RSDP *rsdp;
    Py_ssize_t length = 0;

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    rsdp = (ACPI_TABLE_RSDP *)AcpiOsGetRootPointer();

    if (rsdp) {
        if (rsdp->Revision == 0)
            length = sizeof(ACPI_RSDP_COMMON);
        else if (rsdp->Revision >= 2)
            length = rsdp->Length;
        if (length)
            return Py_BuildValue("s#", rsdp, length);
    }
    return Py_BuildValue("");
}

static PyObject *bits_acpi_get_rsdt(PyObject *self, PyObject *args)
{
    ACPI_TABLE_RSDP *rsdp;
    ACPI_TABLE_RSDT *rsdt;

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    rsdp = (ACPI_TABLE_RSDP *)AcpiOsGetRootPointer();

    if (rsdp) {
        rsdt = (ACPI_TABLE_RSDT *)rsdp->RsdtPhysicalAddress;
        if (rsdt)
            return Py_BuildValue("s#", rsdt, (Py_ssize_t)rsdt->Header.Length);
    }
    return Py_BuildValue("");
}

static PyObject *bits_acpi_get_xsdt(PyObject *self, PyObject *args)
{
    ACPI_TABLE_RSDP *rsdp;
    ACPI_TABLE_XSDT *xsdt;

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    rsdp = (ACPI_TABLE_RSDP *)AcpiOsGetRootPointer();

    if (rsdp) {
        if (rsdp->Revision >= 2) {
            xsdt = (ACPI_TABLE_XSDT *)(U32)rsdp->XsdtPhysicalAddress;
            if (xsdt)
                return Py_BuildValue("s#", xsdt, (Py_ssize_t)xsdt->Header.Length);
        }
    }
    return Py_BuildValue("");
}

static PyObject *bits_acpi_get_table_by_index(PyObject *self, PyObject *args)
{
    U32 index;
    ACPI_TABLE_HEADER *table_header;

    if (!PyArg_ParseTuple(args, "I", &index))
        return NULL;

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    if (ACPI_FAILURE(AcpiGetTableByIndex(index, &table_header)))
        return Py_BuildValue("");

    return Py_BuildValue("s#", table_header, table_header->Length);
}

struct find_processor_context {
    bool init_cpu;
    U32 caps;
    PyObject *cpupath_list;
    PyObject *devpath_list;
};

static grub_err_t call_osc(ACPI_HANDLE cpu_handle, U32 caps)
{
    grub_err_t ret;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT_LIST Params;
    ACPI_OBJECT Obj[4];
    U32 osc_buffer[2];
    U8 OSC_UUID[16] = {0x16, 0xA6, 0x77, 0x40, 0x0C, 0x29, 0xBE, 0x47, 0x9E, 0xBD, 0xD8, 0x70, 0x58, 0x71, 0x39, 0x53};

    /* Initialize the parameter list */
    Params.Count = 4;
    Params.Pointer = Obj;

    /* Initialize the parameter objects */
    // Intel-specific UUID
    Obj[0].Type = ACPI_TYPE_BUFFER;
    Obj[0].Buffer.Length = 16;
    Obj[0].Buffer.Pointer = OSC_UUID;

    // Revision ID
    Obj[1].Type = ACPI_TYPE_INTEGER;
    Obj[1].Integer.Value = 1;

    // Count of DWORDS
    Obj[2].Type = ACPI_TYPE_INTEGER;
    Obj[2].Integer.Value = 2;

    // Capabilities buffer
    Obj[3].Type = ACPI_TYPE_BUFFER;
    Obj[3].Buffer.Length = 8;
    Obj[3].Buffer.Pointer = (U8 *)osc_buffer;

    osc_buffer[0] = 0;
    osc_buffer[1] = caps; // Capabilities DWORD

    Status = AcpiEvaluateObject(cpu_handle, "_OSC", &Params, &Results);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND)
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
    else
        ret = grub_error(GRUB_ERR_IO, "Evaluating _OSC failed (0x%x %s)\n", Status, AcpiFormatException(Status));

    ACPI_FREE(Results.Pointer);

    return ret;
}

static grub_err_t call_pdc(ACPI_HANDLE cpu_handle, U32 caps)
{
    ACPI_STATUS Status;
    ACPI_OBJECT_LIST Params;
    ACPI_OBJECT Obj;
    U32 pdc_buffer[3];

    /* Initialize the parameter list */
    Params.Count = 1;
    Params.Pointer = &Obj;

    /* Initialize the parameter objects */
    Obj.Type = ACPI_TYPE_BUFFER;
    Obj.Buffer.Length = 12;
    Obj.Buffer.Pointer = (UINT8 *)pdc_buffer;

    /* Initialize the DWORD array */
    pdc_buffer[0] = 1; // Revision
    pdc_buffer[1] = 1; // Count
    pdc_buffer[2] = caps; // Capabilities DWORD

    Status = AcpiEvaluateObject(cpu_handle, "_PDC", &Params, NULL);
    if (Status == AE_OK)
        return GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND)
        return grub_error(GRUB_ERR_TEST_FAILURE, "false");
    else
        return grub_error(GRUB_ERR_IO, "Evaluating _PDC failed (0x%x %s)\n", Status, AcpiFormatException(Status));
}

static ACPI_STATUS find_processor(ACPI_HANDLE ObjHandle, UINT32 NestingLevel ACPI_UNUSED_VAR, void *Context, void **ReturnValue ACPI_UNUSED_VAR)
{
    struct find_processor_context *fpc = Context;
    ACPI_BUFFER Path = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    PyObject *cpupath;

    if (!IsEnabledProcessor(ObjHandle))
        goto out;

    if (ACPI_FAILURE(AcpiGetName(ObjHandle, ACPI_FULL_PATHNAME, &Path))) {
        grub_printf("Couldn't get object name\n");
        goto out;
    }

    if (fpc->init_cpu)
        if (call_osc(ObjHandle, fpc->caps) != GRUB_ERR_NONE)
            call_pdc(ObjHandle, fpc->caps);

    cpupath = Py_BuildValue("s", Path.Pointer);
    PyList_Append(fpc->cpupath_list, cpupath);
    Py_XDECREF(cpupath);

out:
    ACPI_FREE(Path.Pointer);
    return AE_OK;
}

static ACPI_STATUS find_processor_dev(ACPI_HANDLE ObjHandle, UINT32 NestingLevel ACPI_UNUSED_VAR, void *Context, void **ReturnValue ACPI_UNUSED_VAR)
{
    struct find_processor_context *fpc = Context;
    ACPI_BUFFER Path = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    PyObject *cpupath;

    if (!IsEnabledProcessorDev(ObjHandle))
        goto out;

    if (ACPI_FAILURE(AcpiGetName(ObjHandle, ACPI_FULL_PATHNAME, &Path))) {
        grub_printf("Couldn't get object name\n");
        goto out;
    }

    if (fpc->init_cpu)
        if (call_osc(ObjHandle, fpc->caps) != GRUB_ERR_NONE)
            call_pdc(ObjHandle, fpc->caps);

    cpupath = Py_BuildValue("s", Path.Pointer);
    PyList_Append(fpc->devpath_list, cpupath);
    Py_XDECREF(cpupath);

out:
    ACPI_FREE(Path.Pointer);
    return AE_OK;
}

static PyObject *bits_acpi_cpupaths(PyObject *self, PyObject *args)
{
    struct find_processor_context fpc = { .init_cpu = false, .caps = 0xfbf, .cpupath_list = NULL };

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    // Before getting any input parameters, change the value of the capabilities DWORD
    // to match the value used in prior ACPI CPU initialization
    if (acpica_cpus_initialized)
        fpc.caps = acpica_cpus_init_caps;

    if (!PyArg_ParseTuple(args, "|I", &fpc.caps))
        return NULL;

    if (acpica_cpus_initialized) {
        if (fpc.caps != acpica_cpus_init_caps)
            return PyErr_Format(PyExc_ValueError, "Attempt to change current Capabilities DWORD from 0x%x to 0x%x; changing capabilities requires ACPI shutdown (acpi_terminate) and restart.",
                                acpica_cpus_init_caps, fpc.caps);
    } else {
        acpica_cpus_initialized = true;
        acpica_cpus_init_caps = fpc.caps;
        fpc.init_cpu = true;
    }

    fpc.cpupath_list = PyList_New(0);

    AcpiWalkNamespace(ACPI_TYPE_PROCESSOR, ACPI_ROOT_OBJECT, ACPI_UINT32_MAX, find_processor, NULL, &fpc, NULL);

    fpc.devpath_list = PyList_New(0);

    AcpiWalkNamespace(ACPI_TYPE_DEVICE, ACPI_ROOT_OBJECT, ACPI_UINT32_MAX, find_processor_dev, NULL, &fpc, NULL);

    return Py_BuildValue("(NN)", fpc.cpupath_list, fpc.devpath_list);
}

struct find_object_context {

    char *needle;
    PyObject *objpath_list;
};

static ACPI_STATUS find_object(ACPI_HANDLE ObjHandle, UINT32 NestingLevel ACPI_UNUSED_VAR, void *Context, void **ReturnValue ACPI_UNUSED_VAR)
{
    struct find_object_context *foc = Context;
    ACPI_BUFFER Path = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    PyObject *objpath;

    if (ACPI_FAILURE(AcpiGetName(ObjHandle, ACPI_FULL_PATHNAME, &Path))) {
        grub_printf("Couldn't get object name\n");
        goto out;
    }

    if (grub_strstr(Path.Pointer, foc->needle) == NULL)
        goto out;

    objpath = Py_BuildValue("s", Path.Pointer);
    PyList_Append(foc->objpath_list, objpath);
    Py_XDECREF(objpath);

out:
    ACPI_FREE(Path.Pointer);
    return AE_OK;
}

static PyObject *bits_acpi_install_interface(PyObject *self, PyObject *args)
{
    char *interface_name;

    if (!PyArg_ParseTuple(args, "s", &interface_name))
        return NULL;

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    if (ACPI_FAILURE(AcpiInstallInterface(interface_name)))
        return PyErr_Format(PyExc_RuntimeError, "AcpiInstallInterface failed.");

    return Py_BuildValue("");
}

static PyObject *bits_acpi_objpaths(PyObject *self, PyObject *args)
{
    struct find_object_context foc = { .objpath_list = NULL };

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    if (!PyArg_ParseTuple(args, "s", &foc.needle))
        return NULL;

    foc.objpath_list = PyList_New(0);

    AcpiWalkNamespace(ACPI_TYPE_ANY, ACPI_ROOT_OBJECT, ACPI_UINT32_MAX, find_object, NULL, &foc, NULL);

    return foc.objpath_list;
}

static PyObject *bits_acpi_remove_interface(PyObject *self, PyObject *args)
{
    char *interface_name;

    if (!PyArg_ParseTuple(args, "s", &interface_name))
        return NULL;

    if (acpica_init() != GRUB_ERR_NONE)
        return PyErr_Format(PyExc_RuntimeError, "ACPICA module failed to initialize.");

    if (ACPI_FAILURE(AcpiRemoveInterface(interface_name)))
        return PyErr_Format(PyExc_RuntimeError, "AcpiRemoveInterface failed.");

    return Py_BuildValue("");
}

void acpica_terminate(void);

static PyObject *bits_acpi_terminate(PyObject *self, PyObject *args)
{
    acpica_terminate();

    return Py_BuildValue("");
}

static PyObject *os_error_with_filename(int errno_val, const char *path)
{
    errno = errno_val;
    PyErr_SetFromErrnoWithFilename(PyExc_OSError, path);
    errno = 0;
    return NULL;
}

static PyObject *bits__getenv(PyObject *self, PyObject *args)
{
    const char *key, *default_value = NULL;
    const char *value;
    if (!PyArg_ParseTuple(args, "s|s:getenv", &key, &default_value))
        return NULL;
    value = grub_env_get(key);
    return Py_BuildValue("s", value ? value : default_value);
}

static PyObject *getenvdict_result;

static int getenvdict_callback(struct grub_env_var *var)
{
    const char *value = var->read_hook ? var->read_hook(var, var->value) : var->value;
    PyDict_SetItem(getenvdict_result, PyString_FromString(var->name), PyString_FromString(value));
    return 0;
}

static PyObject *bits__getenvdict(PyObject *self, PyObject *args)
{
    PyObject *result;
    getenvdict_result = PyDict_New();
    grub_env_iterate(getenvdict_callback);
    result = getenvdict_result;
    getenvdict_result = NULL;
    return result;
}

#define LATENCY_RECENT_COUNT 6
struct latency_bin {
    U64 max;
    U64 total;
    U64 count;
    U32 recent_index;
    U64 recent_absolute[LATENCY_RECENT_COUNT];
};

#define MSR_SMI_COUNT 0x34
static PyObject *bits_smi_latency(PyObject *self, PyObject *args)
{
    U64 test_duration_tscs;
    PyObject *bin_maxes;
    PyObject *bin_obj = NULL;
    PyObject *recent_list = NULL;
    U32 bsp;
    struct msr smi_count1, smi_count2;
    PyObject *smi_count_obj;
    struct latency_bin *bin;
    U32 bin_len;
    U64 test_start;
    U64 tsc1, tsc2;
    U64 current;
    U32 i;
    U64 max = 0;

    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    bsp = bsp_apicid();
    smi_count1.num = smi_count2.num = MSR_SMI_COUNT;

    if (!PyArg_ParseTuple(args, "KO:smi_latency", &test_duration_tscs, &bin_maxes))
        return NULL;
    if (!PySequence_Check(bin_maxes))
        return PyErr_Format(PyExc_TypeError, "expected a sequence");
    bin_len = PySequence_Length(bin_maxes);
    if (bin_len == -1)
        return PyErr_Format(PyExc_ValueError, "failed to get length of sequence");

    bin = grub_zalloc((bin_len + 1) * sizeof(*bin));
    for (i = 0; i < bin_len; i++) {
        PyObject *bin_max_obj = PySequence_GetItem(bin_maxes, i);
        if (PyLong_Check(bin_max_obj)) {
            bin[i].max = PyLong_AsUnsignedLongLong(bin_max_obj);
        } else if (PyInt_Check(bin_max_obj)) {
            bin[i].max = PyInt_AsUnsignedLongLongMask(bin_max_obj);
        } else {
            Py_CLEAR(bin_max_obj);
            PyErr_Format(PyExc_TypeError, "expected an int or long");
            goto err;
        }
        Py_CLEAR(bin_max_obj);
        if (bin[i].max == (U64)-1)
            goto err;
    }
    bin[bin_len].max = ~0ULL;
    bin_len++;

    smp_rdmsr(bsp, &smi_count1);

    for (test_start = tsc1 = rdtsc64(), tsc2 = rdtsc64(); tsc2 - test_start < test_duration_tscs; tsc1 = tsc2, tsc2 = rdtsc64()) {
        current = tsc2 - tsc1;

        for (i = 0; i < bin_len; i++)
            if (current <= bin[i].max) {
                bin[i].count++;
                bin[i].total += current;
                if (bin[i].recent_index != LATENCY_RECENT_COUNT) {
                    bin[i].recent_absolute[bin[i].recent_index] = tsc2;
                    bin[i].recent_index++;
                }
                break;
            }

        if (current > max)
            max = current;
    }

    smp_rdmsr(bsp, &smi_count2);

    bin_obj = PyList_New(bin_len);
    if (!bin_obj)
        goto err;
    for(i = 0; i < bin_len; i++) {
        PyObject *bin_tuple;
        U32 j;
        recent_list = PyList_New(bin[i].recent_index);
        if (!recent_list)
            goto err;
        for (j = 0; j < bin[i].recent_index; j++) {
            PyObject *long_obj = PyLong_FromUnsignedLongLong(bin[i].recent_absolute[j]);
            if (!long_obj)
                goto err;
            PyList_SET_ITEM(recent_list, j, long_obj);
        }
        bin_tuple = Py_BuildValue("KKKN", bin[i].max, bin[i].total, bin[i].count, recent_list);
        if (!bin_tuple)
            goto err;
        PyList_SET_ITEM(bin_obj, i, bin_tuple);
    }

    if (smi_count1.status == 0 && smi_count2.status == 0)
        smi_count_obj = PyLong_FromUnsignedLongLong(smi_count2.value - smi_count1.value);
    else
        smi_count_obj = Py_BuildValue("");

    return Py_BuildValue("KNN", max, smi_count_obj, bin_obj);

err:
    Py_XDECREF(recent_list);
    Py_XDECREF(bin_obj);
    grub_free(bin);
    return NULL;
}

static PyObject *listdir_result;

static int listdir_callback(const char *filename, const struct grub_dirhook_info *info)
{
    PyObject *path;
    if (strcmp(filename, ".") == 0 || strcmp(filename, "..") == 0)
        return 0;
    path = Py_BuildValue("s", filename);
    PyList_Append(listdir_result, path);
    Py_XDECREF(path);
    return 0;
}

static PyObject *bits__listdir(PyObject *self, PyObject *args)
{
    PyObject *result;
    const char *path;
    if (!PyArg_ParseTuple(args, "s", &path))
        return NULL;
    if (!is_directory(path))
        return os_error_with_filename(ENOTDIR, path);

    listdir_result = PyList_New(0);
    iterate_directory(path, listdir_callback);
    result = listdir_result;
    listdir_result = NULL;
    return result;
}

static PyObject *bits__localtime(PyObject *self, PyObject *args)
{
    struct grub_datetime datetime;
    int weekday;
    long int seconds = -1L;

    if (!PyArg_ParseTuple(args, "|I", &seconds))
        return NULL;

    if (seconds == -1L)
        grub_get_datetime(&datetime);
    else
        grub_unixtime2datetime(seconds, &datetime);

    // Get weekday and convert from Sunday=0 to Monday=0
    weekday = (grub_get_weekday(&datetime) + 6) % 7;

    return Py_BuildValue("HBBBBBiii", datetime.year, datetime.month,
                                      datetime.day, datetime.hour,
                                      datetime.minute, datetime.second,
                                      weekday, -1, -1);
}

static char *memory_keywords[] = { "address", "length", "writable", NULL };

static PyObject *bits_memory(PyObject *self, PyObject *args, PyObject *keywds)
{
    unsigned long address;
    Py_ssize_t length;
    PyObject *writable_obj = NULL;
    int writable = 0;

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "kn|O:memory", memory_keywords, &address, &length, &writable_obj))
        return NULL;
    if (writable_obj) {
        writable = PyObject_IsTrue(writable_obj);
        if (writable < 0)
            return NULL;
    }

    return (writable ? PyBuffer_FromReadWriteMemory : PyBuffer_FromMemory)((void *)address, length);
}

static PyObject *bits_memory_addr(PyObject *self, PyObject *args)
{
    PyObject *mem;
    void *addr;
    if (!PyArg_ParseTuple(args, "O!:memory_addr", &PyBuffer_Type, &mem))
        return NULL;
    if (mem->ob_type->tp_as_buffer->bf_getreadbuffer(mem, 0, &addr) == -1)
        return NULL;
    return Py_BuildValue("I", (unsigned int)addr);
}

static PyObject *bits__putenv(PyObject *self, PyObject *args)
{
    const char *key, *value;
    if (!PyArg_ParseTuple(args, "ss:putenv", &key, &value))
        return NULL;
    if (grub_env_set(key, value) != GRUB_ERR_NONE || grub_env_export(key) != GRUB_ERR_NONE)
        return PyErr_SetFromErrno(PyExc_OSError);
    return Py_BuildValue("");
}

static PyObject *bits__stat(PyObject *self, PyObject *args)
{
    const char *path;
    struct stat st;
    if (!PyArg_ParseTuple(args, "s", &path))
        return NULL;
    if (stat(path, &st) < 0)
        return os_error_with_filename(ENOENT, path);
    return Py_BuildValue("(I,K)", st.st_mode, st.st_size);
}

static PyObject *bits__time(PyObject *self, PyObject *args)
{
    return Py_BuildValue("d", grub_get_time_ms() / 1000.0);
}

static PyObject *bits__unsetenv(PyObject *self, PyObject *args)
{
    const char *key;
    if (!PyArg_ParseTuple(args, "s:unsetenv", &key))
        return NULL;
    grub_env_unset(key);
    return Py_BuildValue("");
}

static PyObject *grub_command_callback;

static grub_err_t grub_cmd_pydispatch(grub_command_t cmd, int argc, char **args)
{
    PyObject *pyargs, *pyret;
    grub_err_t ret;
    unsigned ndx;

    pyargs = PyList_New(argc+1);
    if (!pyargs)
        return GRUB_ERR_OUT_OF_MEMORY;

    PyList_SET_ITEM(pyargs, 0, PyString_FromString(cmd->name));
    for (ndx = 0; ndx < argc; ndx++)
        PyList_SET_ITEM(pyargs, ndx+1, PyString_FromString(args[ndx]));

    pyret = PyObject_CallFunctionObjArgs(grub_command_callback, pyargs, NULL);
    Py_DECREF(pyargs);

    if (pyret == NULL)
        return grub_error(GRUB_ERR_IO, "Internal error: Failed to call Python command callback, or it threw an exception");

    if (pyret == Py_None || PyObject_IsTrue(pyret))
        ret = GRUB_ERR_NONE;
    else
        ret = GRUB_ERR_TEST_FAILURE;
    Py_DECREF(pyret);

    return ret;
}

static PyObject *bits_register_grub_command(PyObject *self, PyObject *args)
{
    const char *cmd, *summary, *description;
    char *cmd_copy, *summary_copy, *description_copy;

    if (!grub_command_callback)
        return PyErr_Format(PyExc_RuntimeError, "Internal error: attempted to register grub command before setting callback.");

    if (!PyArg_ParseTuple(args, "sss:register_grub_command", &cmd, &summary, &description))
        return NULL;

    cmd_copy = grub_strdup(cmd);
    summary_copy = grub_strdup(summary);
    description_copy = grub_strdup(description);
    if (!cmd_copy || !summary_copy || !description_copy) {
        grub_free(cmd_copy);
        grub_free(summary_copy);
        grub_free(description_copy);
        return PyErr_NoMemory();
    }

    grub_register_command(cmd_copy, grub_cmd_pydispatch, summary_copy, description_copy);

    return Py_BuildValue("");
}

static PyObject *bits_set_grub_command_callback(PyObject *self, PyObject *args)
{
    PyObject *callable;
    if (!PyArg_ParseTuple(args, "O:set_grub_command_callback", &callable))
        return NULL;

    if (!PyCallable_Check(callable))
        return PyErr_Format(PyExc_TypeError, "expected a callable");

    Py_XDECREF(grub_command_callback);
    Py_XINCREF(callable);
    grub_command_callback = callable;

    return Py_BuildValue("");
}

static PyObject *readline_callback;

/* Note that this always assumes in and out are sys.stdin and sys.stdout. */
static char *bits_readline_function(FILE *in, FILE *out, char *prompt)
{
    PyObject *pyret;
    Py_ssize_t len;
    char *temp;
    char *ret;

    (void)in;
    (void)out;

    pyret = PyObject_CallFunction(readline_callback, "s", prompt);

    if (!pyret)
        return NULL;
    if (!PyString_Check(pyret)) {
        PyErr_Format(PyExc_TypeError, "Python readline callback returned a non-string");
        return NULL;
    }
    temp = PyString_AsString(pyret);
    if (!temp)
        return NULL;
    len = PyString_Size(pyret);
    ret = PyMem_Malloc(len+1);
    if (!ret) {
        PyErr_NoMemory();
        return NULL;
    }
    return memcpy(ret, temp, len+1);
}

static PyObject *bits_set_readline_callback(PyObject *self, PyObject *args)
{
    PyObject *callable;
    if (!PyArg_ParseTuple(args, "O:set_readline_callback", &callable))
        return NULL;

    if (callable == Py_None) {
        PyOS_ReadlineFunctionPointer = NULL;
        Py_XDECREF(readline_callback);
        readline_callback = Py_BuildValue("");
        return Py_BuildValue("");
    }

    if (!PyCallable_Check(callable))
        return PyErr_Format(PyExc_TypeError, "expected a callable");

    Py_XDECREF(readline_callback);
    Py_XINCREF(callable);
    readline_callback = callable;
    PyOS_ReadlineFunctionPointer = bits_readline_function;

    return Py_BuildValue("");
}

static PyObject *bits_get_key(PyObject *self, PyObject *args)
{
    return Py_BuildValue("i", grub_getkey());
}

static PyObject *bits_get_mwait(PyObject *self, PyObject *args)
{
    U32 apicid;
    bool use_mwait;
    U32 hint, int_break_event;
    if (!PyArg_ParseTuple(args, "I:get_mwait", &apicid))
        return NULL;
    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    if (!smp_get_mwait(apicid, &use_mwait, &hint, &int_break_event))
        return PyErr_Format(PyExc_RuntimeError, "Failed to get mwait hint for apicid %u", apicid);
    return Py_BuildValue("NIN", PyBool_FromLong(use_mwait), hint, PyBool_FromLong(int_break_event));
}

static PyObject *bits_set_mwait(PyObject *self, PyObject *args)
{
    U32 apicid, hint = 0;
    PyObject *use_mwait_obj, *int_break_event_obj = NULL;
    if (!PyArg_ParseTuple(args, "IO|IO:set_mwait", &apicid, &use_mwait_obj, &hint, &int_break_event_obj))
        return NULL;
    if (!smp_init())
        return PyErr_Format(PyExc_RuntimeError, "SMP module failed to initialize.");
    smp_set_mwait(apicid, PyObject_IsTrue(use_mwait_obj), hint, int_break_event_obj ? PyObject_IsTrue(int_break_event_obj) : 1);
    return Py_BuildValue("");
}

static PyObject *bits_clear_screen(PyObject *self, PyObject *args)
{
    grub_cls();
    return Py_BuildValue("");
}

/* GRUB keeps terminals in a linked list, and this family of functions exposes
 * them via an index.  That makes the resulting Python code O(N^2), but since N
 * is generally no more than 2... */
static PyObject *bits_get_term_count(PyObject *self, PyObject *args)
{
    grub_term_output_t term;
    unsigned count = 0;

    FOR_ACTIVE_TERM_OUTPUTS(term)
        count++;

    return Py_BuildValue("I", count);
}

static char *term_keyword[] = {"term", NULL};

static PyObject *bits_get_width_height(PyObject *self, PyObject *args, PyObject *keywds)
{
    unsigned term_num = 0;
    unsigned current = 0;
    grub_term_output_t term;

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "I", term_keyword, &term_num))
        return NULL;

    FOR_ACTIVE_TERM_OUTPUTS(term) {
        if (current == term_num) {
            U16 temp = term->getwh(term);
            return Py_BuildValue("BB", (temp >> 8) & 0xff, temp & 0xff);
        }
        current++;
    }

    return PyErr_Format(PyExc_ValueError, "term (%u) must be less than %u.", term_num, current);
}

static PyObject *bits_get_xy(PyObject *self, PyObject *args, PyObject *keywds)
{
    unsigned term_num;
    unsigned current = 0;
    grub_term_output_t term;

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "I", term_keyword, &term_num))
        return NULL;

    FOR_ACTIVE_TERM_OUTPUTS(term) {
        if (current == term_num) {
            U16 temp = term->getxy(term);
            return Py_BuildValue("BB", (temp >> 8) & 0xff, temp & 0xff);
        }
        current++;
    }

    return PyErr_Format(PyExc_ValueError, "term (%u) must be less than %u.", term_num, current);
}

static char *xy_keywords[] = {"x", "y", "term", NULL};

static PyObject *bits_goto_xy(PyObject *self, PyObject *args, PyObject *keywds)
{
    unsigned term_num;
    unsigned current = 0;
    grub_term_output_t term;
    U8 x, y;

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "BBI", xy_keywords, &x, &y, &term_num))
        return NULL;

    FOR_ACTIVE_TERM_OUTPUTS(term) {
        if (current == term_num) {
            term->gotoxy(term, x, y);
            return Py_BuildValue("");
        }
        current++;
    }

    return PyErr_Format(PyExc_ValueError, "term (%u) must be less than %u.", term_num, current);
}

static char *puts_keywords[] = {"str", "term", NULL};

static PyObject *bits_puts(PyObject *self, PyObject *args, PyObject *keywds)
{
    unsigned term_num;
    unsigned current = 0;
    grub_term_output_t term;
    char *str;

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "sI", puts_keywords, &str, &term_num))
        return NULL;

    FOR_ACTIVE_TERM_OUTPUTS(term) {
        if (current == term_num) {
            grub_puts_terminal(str, term);
            return Py_BuildValue("");
        }
        current++;
    }

    return PyErr_Format(PyExc_ValueError, "term (%u) must be less than %u.", term_num, current);
}

static PyObject *pyfs_dir_callable;
static PyObject *pyfs_open_callable;
static PyObject *pyfs_read_callable;

grub_err_t do_pyfs_dir(const char *path, int (*hook)(const char *filename, const struct grub_dirhook_info *info));
grub_err_t do_pyfs_dir(const char *path, int (*hook)(const char *filename, const struct grub_dirhook_info *info))
{
    PyObject *pyret;
    PyObject *iterator;
    PyObject *item;

    if (!pyfs_dir_callable)
        return GRUB_ERR_FILE_NOT_FOUND;

    pyret = PyObject_CallFunction(pyfs_dir_callable, "s", path);

    if (pyret == NULL) {
        PyErr_Print();
        return grub_error(GRUB_ERR_IO, "Internal error: Failed to call Python dir callback, or it threw an exception");
    }

    if (pyret == Py_None) {
        Py_DECREF(pyret);
        return GRUB_ERR_BAD_FILE_TYPE;
    }

    iterator = PyObject_GetIter(pyret);
    Py_DECREF(pyret);
    if (!iterator) {
        PyErr_Print();
        return grub_error(GRUB_ERR_IO, "Internal error: Python dir callback did not return a sequence");
    }

    while ((item = PyIter_Next(iterator))) {
        char *pyname;
        PyObject *pyisdir;
        if (!PyArg_ParseTuple(item, "sO", &pyname, &pyisdir)) {
            break;
        }

        {
            struct grub_dirhook_info info = {
                .dir = PyObject_IsTrue(pyisdir)
            };
            if (hook(pyname, &info))
                break;
        }
        Py_DECREF(item);
    }
    Py_XDECREF(item); /* If we broke out of the loop, handle the last item. */

    Py_DECREF(iterator);
    if (PyErr_Occurred()) {
        PyErr_Print();
        return grub_error(GRUB_ERR_IO, "Internal error: Python dir callback produced an error while iterating");
    }

    return GRUB_ERR_NONE;
}

grub_err_t do_pyfs_open(const char *name, grub_off_t *size);
grub_err_t do_pyfs_open(const char *name, grub_off_t *size)
{
    PyObject *pyret;
    Py_ssize_t pysize;

    if (!pyfs_open_callable)
        return GRUB_ERR_FILE_NOT_FOUND;

    pyret = PyObject_CallFunction(pyfs_open_callable, "s", name);

    if (pyret == NULL) {
        PyErr_Print();
        return grub_error(GRUB_ERR_IO, "Internal error: Failed to call Python open callback, or it threw an exception");
    }

    if (pyret == Py_None) {
        Py_DECREF(pyret);
        return GRUB_ERR_BAD_FILE_TYPE;
    }

    pysize = PyInt_AsSsize_t(pyret);
    Py_DECREF(pyret);
    if (pysize < 0) {
        if (PyErr_Occurred())
            PyErr_Print();
        return grub_error(GRUB_ERR_IO, "Internal error: Python open callback returned a bad or negative size");
    }

    *size = (grub_off_t)pysize;
    return GRUB_ERR_NONE;
}

grub_ssize_t do_pyfs_read(const char *name, grub_off_t offset, void *buf, grub_size_t len);
grub_ssize_t do_pyfs_read(const char *name, grub_off_t offset, void *buf, grub_size_t len)
{
    PyObject *pyret;
    char *pybuf;
    Py_ssize_t pylen;

    if (!pyfs_read_callable)
        return -1;

    pyret = PyObject_CallFunction(pyfs_read_callable, "sKK", name, (unsigned long long)offset, (unsigned long long)len);

    if (pyret == NULL) {
        PyErr_Print();
        grub_error(GRUB_ERR_IO, "Internal error: Failed to call Python read callback, or it threw an exception");
        return -1;
    }
    if (!PyString_Check(pyret) || PyString_AsStringAndSize(pyret, &pybuf, &pylen) < 0) {
        if (PyErr_Occurred())
            PyErr_Print();
        Py_DECREF(pyret);
        grub_error(GRUB_ERR_IO, "Internal error: Python read callback returned a bad string");
        return -1;
    }
    if ((grub_off_t)pylen != len) {
        Py_DECREF(pyret);
        grub_error(GRUB_ERR_IO, "Internal error: Expected %llu bytes but Python read callback returned %llu", (unsigned long long)len, (unsigned long long)pylen);
        return -1;
    }

    grub_memcpy(buf, pybuf, pylen);
    Py_DECREF(pyret);
    return pylen;
}

static PyObject *bits_set_pyfs_callbacks(PyObject *self, PyObject *args)
{
    PyObject *pyfs_dir_callable_temp, *pyfs_open_callable_temp, *pyfs_read_callable_temp;
    if (!PyArg_ParseTuple(args, "OOO:_set_pyfs_callbacks", &pyfs_dir_callable_temp, &pyfs_open_callable_temp, &pyfs_read_callable_temp))
        return NULL;

    if (!PyCallable_Check(pyfs_dir_callable_temp))
        return PyErr_Format(PyExc_TypeError, "expected a callable for pyfs_dir");
    if (!PyCallable_Check(pyfs_open_callable_temp))
        return PyErr_Format(PyExc_TypeError, "expected a callable for pyfs_open");
    if (!PyCallable_Check(pyfs_read_callable_temp))
        return PyErr_Format(PyExc_TypeError, "expected a callable for pyfs_read");

    Py_XDECREF(pyfs_dir_callable);
    Py_XINCREF(pyfs_dir_callable_temp);
    pyfs_dir_callable = pyfs_dir_callable_temp;

    Py_XDECREF(pyfs_open_callable);
    Py_XINCREF(pyfs_open_callable_temp);
    pyfs_open_callable = pyfs_open_callable_temp;

    Py_XDECREF(pyfs_read_callable);
    Py_XINCREF(pyfs_read_callable_temp);
    pyfs_read_callable = pyfs_read_callable_temp;

    return Py_BuildValue("");
}

PyDoc_STRVAR(bits_set_pyfs_callbacks_doc,
"_set_pyfs_callbacks(pyfs_dir, pyfs_open, pyfs_read)\n"
"\n"
"Set the callbacks implementing the (python) filesystem.\n"
"These callbacks should be callables with the following signatures:\n"
"\n"
"pyfs_dir(dirname):\n"
"    return an iterable of (filename, is_directory) pairs, or None if\n"
"    not a directory\n"
"pyfs_open(filename):\n"
"    return the file size, or None if the file does not exist\n"
"pyfs_read(filename, offset, size):\n"
"    return size bytes starting at offset, as a string\n"
);

static PyMethodDef bitsMethods[] = {
    {"_acpi_cpupaths",  bits_acpi_cpupaths, METH_VARARGS, "_acpi_cpupaths([capabilities]) -> tuple(list of cpu namepaths, list of device namepaths)"},
    {"_acpi_eval",  bits_acpi_eval, METH_VARARGS, "_acpi_eval(\"\\PATH._TO_.EVAL\") -> result"},
    {"_acpi_get_object_info",  bits_acpi_get_object_info, METH_VARARGS, "_acpi_get_object_info() -> (infostr, address)"},
    {"_acpi_get_rsdp",  bits_acpi_get_rsdp, METH_NOARGS, "_acpi_get_rsdp() -> str"},
    {"_acpi_get_rsdt",  bits_acpi_get_rsdt, METH_NOARGS, "_acpi_get_rsdt() -> str"},
    {"_acpi_get_xsdt",  bits_acpi_get_xsdt, METH_NOARGS, "_acpi_get_xsdt() -> str"},
    {"_acpi_get_table",  bits_acpi_get_table, METH_VARARGS, "_acpi_get_table(signature[, instance=1]) -> str"},
    {"_acpi_get_table_by_index",  bits_acpi_get_table_by_index, METH_VARARGS, "_acpi_get_table_by_index(index) -> str"},
    {"_acpi_install_interface", bits_acpi_install_interface, METH_VARARGS, "_acpi_install_interface(\"interface_name\")"},
    {"_acpi_objpaths",  bits_acpi_objpaths, METH_VARARGS, "_acpi_objpaths(\"objectname\") -> list of obj namepaths"},
    {"_acpi_remove_interface", bits_acpi_remove_interface, METH_VARARGS, "_acpi_remove_interface(\"interface_name\")"},
    {"_acpi_terminate", bits_acpi_terminate, METH_NOARGS, "_acpi_terminate() -> Perform ACPICA module terminate"},
    {"bclk", bits_bclk, METH_NOARGS, "bclk() -> bclk (in MHz)"},
    {"blocking_sleep", bits_blocking_sleep, METH_VARARGS, "sleep using mwait for the specified number of microseconds"},
    {"clear_screen", bits_clear_screen, METH_NOARGS, "clear_screen() -> clear the screen"},
    {"_cpuid", bits_cpuid, METH_VARARGS, "_cpuid(apicid, eax[, ecx]) -> eax, ebx, ecx, edx"},
    {"cpus",  bits_cpus, METH_NOARGS, "cpus() -> list of APIC IDs"},
    {"disk_read", (PyCFunction)bits_disk_read, METH_VARARGS, "disk_read(file, sector, offset, length) -> data. Uses file to identify disk."},
    {"disk_write", (PyCFunction)bits_disk_write, METH_VARARGS, "disk_write(file, sector, offset, data). Uses file to identify disk."},
    {"file_data_and_disk_blocks", (PyCFunction)bits_file_data_and_disk_blocks, METH_VARARGS, "file_data_and_disk_blocks(file) -> (data, [(sector, offset, length), ...])"},
    {"_getenv",  bits__getenv, METH_VARARGS, "_getenv(key, default=None) -> value of environment variable \"key\", or default if it doesn't exist"},
    {"_getenvdict",  bits__getenvdict, METH_NOARGS, "_getenvdict() -> environment dictionary"},
    {"get_key", bits_get_key, METH_NOARGS, "get_key() -> keycode"},
    {"get_mwait", bits_get_mwait, METH_VARARGS, "get_mwait(apicid) -> (use_mwait, hint, int_break_event)"},
    {"get_term_count", bits_get_term_count, METH_NOARGS, "get_term_count() -> number of terminals"},
    {"get_width_height", (PyCFunction)bits_get_width_height, METH_KEYWORDS, "get_width_height(term) -> (width, height)" },
    {"get_xy", (PyCFunction)bits_get_xy, METH_KEYWORDS, "get_xy(term) -> (cursor_x, cursor_y)"},
    {"goto_xy", (PyCFunction)bits_goto_xy, METH_KEYWORDS, "goto_xy(x, y, term)) -> position cursor at these coordinates"},
    {"inb", (PyCFunction)bits_inb, METH_KEYWORDS, "inb(port[, apicid=BSP]) -> read byte from IO port on the specified CPU"},
    {"inw", (PyCFunction)bits_inw, METH_KEYWORDS, "inw(port[, apicid=BSP]) -> read word from IO port on the specified CPU"},
    {"inl", (PyCFunction)bits_inl, METH_KEYWORDS, "inl(port[, apicid=BSP]) -> read dword from IO port on the specified CPU"},
    {"_listdir",  bits__listdir, METH_VARARGS, "_listdir() -> list of pathnames"},
    {"_localtime", bits__localtime, METH_VARARGS, "_localtime([seconds]) -> tuple (internal implementation details of localtime)"},
    {"memory", (PyCFunction)bits_memory, METH_KEYWORDS, "memory(address, length[, writable=False]) -> buffer"},
    {"memory_addr", bits_memory_addr, METH_VARARGS, "memory_addr(mem) -> address of mem, which must have been returned by bits.memory"},
    {"outb", (PyCFunction)bits_outb, METH_KEYWORDS, "outb(port, value[, apicid=BSP]) -> write byte to IO port on the specified CPU"},
    {"outw", (PyCFunction)bits_outw, METH_KEYWORDS, "outw(port, value[, apicid=BSP]) -> write word to IO port on the specified CPU"},
    {"outl", (PyCFunction)bits_outl, METH_KEYWORDS, "outl(port, value[, apicid=BSP]) -> write dword to IO port on the specified CPU"},
    {"puts", (PyCFunction)bits_puts, METH_KEYWORDS, "puts(string, term)) -> puts string to specified terminal"},
    {"_putenv",  bits__putenv, METH_VARARGS, "_putenv(key, value): Set an environment variable"},
    {"rdmsr",  bits_rdmsr, METH_VARARGS, "rdmsr(apicid, msr) -> long (None if GPF)"},
    {"readb", (PyCFunction)bits_readb, METH_KEYWORDS, "readb(address[, apicid=BSP]) -> read byte from memory on the specified CPU"},
    {"readw", (PyCFunction)bits_readw, METH_KEYWORDS, "readw(address[, apicid=BSP]) -> read word from memory on the specified CPU"},
    {"readl", (PyCFunction)bits_readl, METH_KEYWORDS, "readl(address[, apicid=BSP]) -> read dword from memory on the specified CPU"},
    {"readq", (PyCFunction)bits_readq, METH_KEYWORDS, "readq(address[, apicid=BSP]) -> read qword from memory on the specified CPU"},
    {"_register_grub_command", bits_register_grub_command, METH_VARARGS, "register_grub_command(name, summary, description)"},
    {"_set_grub_command_callback", bits_set_grub_command_callback, METH_VARARGS, "set_grub_command_callback(callable)"},
    {"_set_readline_callback", bits_set_readline_callback, METH_VARARGS, "_set_readline_callback(callable)"},
    {"_set_pyfs_callbacks", bits_set_pyfs_callbacks, METH_VARARGS, bits_set_pyfs_callbacks_doc},
    {"set_mwait", bits_set_mwait, METH_VARARGS, "set_mwait(apicid, use_mwait[, hint=0[, int_break_event=True]]) -> Enable/disable MWAIT, and set hints and flags"},
    {"_stat", bits__stat, METH_VARARGS, "_stat(path) -> tuple (internal implementation details of stat)"},
    {"smi_latency", bits_smi_latency, METH_VARARGS, "smi_latency(duration, bin_maxes) -> (max_latency, smi_count_delta, [(bin_max, bin_total, bin_count, [latency])]). All times in TSC counts. smi_count_delta is None if reading MSR_SMI_COUNT GPFs."},
    {"_time", bits__time, METH_NOARGS, "_time() -> time in seconds (accurate for relative use only)"},
    {"_unsetenv",  bits__unsetenv, METH_VARARGS, "_unsetenv(key): Unset an environment variable"},
    {"writeb", (PyCFunction)bits_writeb, METH_KEYWORDS, "writeb(address, value[, apicid=BSP]) -> write byte to memory on the specified CPU"},
    {"writew", (PyCFunction)bits_writew, METH_KEYWORDS, "writew(address, value[, apicid=BSP]) -> write word to memory on the specified CPU"},
    {"writel", (PyCFunction)bits_writel, METH_KEYWORDS, "writel(address, value[, apicid=BSP]) -> write dword to memory on the specified CPU"},
    {"writeq", (PyCFunction)bits_writeq, METH_KEYWORDS, "writeq(address, value[, apicid=BSP]) -> write qword to memory on the specified CPU"},
    {"wrmsr",  bits_wrmsr, METH_VARARGS, "wrmsr(apicid, msr, value) -> bool (False if GPF, True otherwise)"},
    {NULL, NULL, 0, NULL}        /* Sentinel */
};

PyMODINIT_FUNC init_bits(void)
{
    (void) Py_InitModule("_bits", bitsMethods);
}
